from django.utils.translation import gettext_lazy as _l

from pymess.utils import fullname


class BaseBackend:

    model = None

    def __init__(self):
        assert self.model is not None, _l('self.model cannot be None')

    def _get_extra_sender_data(self):
        """
        Gets arguments that will be saved with the message in the extra_sender_data field
        """
        return {}

    def _get_extra_message_kwargs(self):
        """
        Gets model message kwargs that will be saved with the message
        """
        return {}

    def update_message(self, message, extra_sender_data=None, **kwargs):
        """
        Method for updating state of the message after sending
        :param message: message object
        :param extra_sender_data: extra data that will be saved to the extra_sender_data field
        :param kwargs: changed object kwargs
        :return:
        """
        extra_sender_data = {
            **self._get_extra_sender_data(),
            **({} if extra_sender_data is None else extra_sender_data)
        }
        message.change_and_save(
            backend=fullname(self),
            extra_sender_data=extra_sender_data,
            **kwargs
        )

    def create_message(self, recipient, content, related_objects, tag, template, **kwargs):
        """
        Create message which will be logged in the database.
        :param recipient: email or phone number of the recipient
        :param content: content of the message
        :param related_objects: list of related objects that will be linked with the message using generic
        relation
        :param tag: string mark that will be saved with the message
        :param template: template object from which content of the message was created
        :param kwargs: extra attributes that will be saved with the message
        """
        message = self.model.objects.create(
            recipient=recipient,
            content=content,
            tag=tag,
            template=template,
            template_slug=template.slug if template else None,
            **kwargs
        )
        if related_objects:
            message.related_objects.create_from_related_objects(*related_objects)
        return message

    def publish_message(self, message):
        """
        Send the message
        :param message: SMS message
        """
        raise NotImplementedError

    def publish_messages(self, messages):
        """
        Sends more messages together. If concrete backend provides send more messages at once the method
        can be overridden
        :param messages: list of messages
        """
        [self.publish_message(message) for message in messages]

    def send(self, recipient, content, related_objects=None, tag=None, template=None, **kwargs):
        """
        Send message with the text content to the phone number (recipient)
        :param recipient: email or phone number of the recipient
        :param content: text content of the message
        :param related_objects: list of related objects that will be linked with the message using generic
        relation
        :param tag: string mark that will be saved with the message
        :param template: template object from which content of the message was create
        :param kwargs: extra attributes that will be stored to the message
        """
        message = self.create_message(recipient, content, related_objects, tag, template, **kwargs)
        self.publish_message(message)
        return message

    def bulk_send(self, recipients, content, related_objects=None, tag=None, template=None, **kwargs):
        """
        Send more messages in one bulk
        :param recipients: list of emails or phone numbers of recipients
        :param content: text content of the messages
        :param related_objects: list of related objects that will be linked with the message using generic
        relation
        :param tag: string mark that will be saved with the message
        :param template: template object from which content of the message was create
        :param kwargs: extra attributes that will be stored with messages
        """
        messages = [
            self.create_message(recipient, content, related_objects, tag, template, **kwargs)
            for recipient in recipients
        ]
        self.publish_messages(messages)
        return messages


def send_template(recipient, slug, context_data, related_objects=None, tag=None, template_model=None, **kwargs):
    """
    Helper for building and sending message from a template.
    :param recipient: email or phone number of the recipient
    :param slug: slug of a template
    :param context_data: dict of data that will be sent to the template renderer
    :param related_objects: list of related objects that will be linked with the dialer message using generic
        relation
    :param tag: string mark that will be saved with the message
    :param template_model: template model instance
    :return: dialer message object or None if template cannot be sent
    """

    assert template_model is not None, _l('template_model cannot be None')

    return template_model.objects.get(slug=slug).send(
        recipient,
        context_data,
        related_objects=related_objects,
        tag=tag,
        **kwargs
    )


def send(recipient, content, related_objects=None, tag=None, message_sender=None, **kwargs):
    """
    Helper for sending message.
    :param recipient: email or phone number of the recipient
    :param content: text content of the messages
    :param related_objects:
    :param tag: string mark that will be saved with the message
    :param kwargs: extra attributes that will be stored with messages
    :param message_sender: sender instance 
    :return: True if message was successfully sent or False if message is in error state
    """
    return message_sender.send(
        recipient,
        content,
        related_objects=related_objects,
        tag=tag,
        **kwargs
    ).failed



from pymess.backend.dialer import DialerBackend
from pymess.models import DialerMessage


class DummyDialerBackend(DialerBackend):
    """
    Dummy dialer backend used for testing environments. Backend only logs messages to the database.
    """

    def publish_message(self, message):
        self.update_message(message, state=DialerMessage.STATE.DEBUG)



from datetime import timedelta

from django.db.models import Q
from django.core.management.base import BaseCommand, CommandError
from django.utils.timezone import now

from pymess.config import get_email_sender, settings
from pymess.lockfile import FileLock, AlreadyLocked, LockTimeout
from pymess.models import EmailMessage


class Command(BaseCommand):
    """
    Command for sending e-mails in batch.
    """

    def handle(self, *args, **options):
        if not settings.EMAIL_BATCH_SENDING:
            raise CommandError('Batch sending is turned off')

        lock = FileLock(settings.EMAIL_BATCH_LOCK_FILE)
        try:
            self.stdout.write('acquiring lock...')
            lock.acquire(settings.EMAIL_BATCH_LOCK_WAIT_TIMEOUT)
        except AlreadyLocked:
            raise CommandError('lock already in place. quitting.')
        except LockTimeout:
            raise CommandError('waiting for the lock timed out. quitting.')

        email_sender = get_email_sender()

        sent = 0
        try:
            messages_to_send = EmailMessage.objects.filter(
                Q(state=EmailMessage.STATE.WAITING) |
                Q(
                    state=EmailMessage.STATE.ERROR,
                    number_of_send_attempts__lte=settings.EMAIL_BATCH_MAX_NUMBER_OF_SEND_ATTEMPTS,
                    created_at__gte=now() - timedelta(seconds=settings.EMAIL_BATCH_MAX_SECONDS_TO_SEND)
                )
            ).order_by('created_at')[:settings.EMAIL_BATCH_SIZE]
            for message in messages_to_send:
                email_sender.publish_message(message)
                sent += 1
            self.stdout.write('{} e-mails sent'.format(sent))
        finally:
            self.stdout.write('releasing lock...')
            lock.release()
            self.stdout.write('released.')



from chamber.exceptions import PersistenceException
from django.utils.encoding import force_text

from pymess.backend import BaseBackend
from pymess.models import PushNotificationMessage


class PushNotificationBackend(BaseBackend):
    """Base class for push notification backend with implementation of push notification service used for sending."""

    model = PushNotificationMessage

    class PushNotificationSendingError(Exception):
        pass

    def create_message(self, recipient, content, related_objects, tag, template, **kwargs):
        try:
            notification = super().create_message(
                recipient=recipient,
                content=content,
                related_objects=related_objects,
                template=template,
                tag=tag,
                **kwargs
            )
            return notification
        except PersistenceException as ex:
            raise self.PushNotificationSendingError(force_text(ex))



from django.utils import timezone as tz
from django.utils.encoding import force_text

from pymess.backend.dialer import DialerBackend
from pymess.config import settings
from pymess.models import DialerMessage
from pymess.utils.logged_requests import generate_session


class DaktelaDialerBackend(DialerBackend):
    """
    Dialer backend implementing Daktela service https://www.daktela.com/api/v6/models/campaignsrecords
    """

    SESSION_SLUG = 'pymess-daktela_autodialer'

    @staticmethod
    def _get_dialer_api_url(name=None):
        name = '/{name}'.format(name=name) if name else ''

        return '{base_url}{name}.json?accessToken={access_token}'.format(
            base_url=settings.DIALER_DAKTELA.URL, name=name, access_token=settings.DIALER_DAKTELA.ACCESS_TOKEN,
        )

    def _update_dialer_states(self, messages):
        """
        Method uses Daktela API to get info about autodialer call status
        :param messages: list of dialer messages to update
        """
        for message in messages:
            name = message.extra_data['name']
            client_url = self._get_dialer_api_url(name)
            response = generate_session(slug=self.SESSION_SLUG, related_objects=(message,)).get(client_url)
            resp_json = response.json()

            message.extra_data.update({
                'name': name,
                'daktela_action': resp_json['result']['action'],
                'daktela_statuses': resp_json['result']['statuses'],
            })
            resp_message_state = resp_json['result']['statuses'][0]['name'] if len(
                resp_json['result']['statuses']) else resp_json['result']['action']
            custom_fields = resp_json['result']['customFields']
            tts_processed = custom_fields['ttsprocessed'][0]
            whole_message_heard = ('whole_message_heard' in custom_fields and custom_fields['whole_message_heard']
                                   and custom_fields['whole_message_heard'][0] == 'Yes')
            state_mapped = settings.DIALER_DAKTELA.STATES_MAPPING[resp_message_state]
            if state_mapped == DialerMessage.STATE.ANSWERED_PARTIAL and whole_message_heard:
                resp_message_state = str(DialerMessage.STATE.ANSWERED_COMPLETE)
            if state_mapped == DialerMessage.STATE.DONE and tts_processed == '0':
                resp_message_state = str(DialerMessage.STATE.NOT_ASSIGNED)
            message_state = settings.DIALER_DAKTELA.STATES_MAPPING[resp_message_state]
            message_error = resp_json['error'] if len(resp_json['error']) else None
            tts_processed = resp_json['result']['customFields']['ttsprocessed'][0]

            try:
                self.update_message(
                    message,
                    state=message_state,
                    error=message_error,
                    extra_data=message.extra_data,
                    is_final_state=resp_json['result']['action'] == '5' and tts_processed == '1',
                )
            except Exception as ex:
                self.update_message(message, state=DialerMessage.STATE.ERROR, error=force_text(ex))
                # Do not re-raise caught exception. We do not know exact exception to catch so we catch them all
                # and log them into database. Re-raise exception causes transaction rollback (lost of information about
                # exception).

    def publish_message(self, message):
        """
        Method uses Daktela API for sending dialer message
        :param message: dialer message
        """
        client_url = self._get_dialer_api_url()
        try:
            payload = {
                'queue': settings.DIALER_DAKTELA.QUEUE,
                'number': message.recipient,
                'customFields': {
                    'mall_pay_text': [
                        message.content,
                    ],
                    'ttsprocessed': [
                        0,
                    ],
                },
                'action': 5,
            }

            response = generate_session(slug=self.SESSION_SLUG, related_objects=(message,)).post(
                client_url,
                json=payload,
            )
            resp_json = response.json()

            message.extra_data.update({
                'name': resp_json['result']['name'],
                'daktela_action': resp_json['result']['action'],
                'daktela_statuses': resp_json['result']['statuses'],
            })

            self.update_message(
                message,
                state=DialerMessage.STATE.READY,
                error=resp_json['error'] if len(resp_json['error']) else None,
                sent_at=tz.now(),
                extra_data=message.extra_data,
            )
        except Exception as ex:
            self.update_message(message, state=DialerMessage.STATE.ERROR, error=force_text(ex))
            # Do not re-raise caught exception. We do not know exact exception to catch so we catch them all
            # and log them into database. Re-raise exception causes transaction rollback (lost of information about
            # exception).



from http import HTTPStatus
from json.decoder import JSONDecodeError

import requests
from django.utils import timezone
from django.utils.encoding import force_text
from onesignal import DeviceNotification, OneSignalClient

from pymess.backend.push import PushNotificationBackend
from pymess.config import settings
from pymess.models import PushNotificationMessage
from pymess.utils.logged_requests import generate_session


class OneSignalPushNotificationBackend(PushNotificationBackend):

    def _is_result_partial_error(self, result):
        return not result.is_error and result.errors

    def _is_invalid_result(self, result):
        return result.is_error or self._is_result_partial_error(result)

    def publish_message(self, message):
        onesignal_client = OneSignalClient(settings.PUSH_NOTIFICATION_ONESIGNAL_APP_ID,
                                           settings.PUSH_NOTIFICATION_ONESIGNAL_API_KEY)
        onesignal_client.session = generate_session(slug='pymess - OneSignal', related_objects=(message,))

        languages = {'en'}
        if settings.PUSH_NOTIFICATION_ONESIGNAL_LANGUAGE is not None:
            languages.add(settings.PUSH_NOTIFICATION_ONESIGNAL_LANGUAGE)
        notification = DeviceNotification(
            include_external_user_ids=(message.recipient,),
            contents={language: message.content for language in languages},
            headings={language: message.heading for language in languages},
            data=message.extra_data,
            url=message.url,
        )

        try:
            result = onesignal_client.send(notification)

            extra_sender_data = message.extra_sender_data or {}
            extra_sender_data['result'] = result.body

            error_state, sent_state = PushNotificationMessage.STATE.ERROR, PushNotificationMessage.STATE.SENT
            self.update_message(message, state=error_state if self._is_invalid_result(result) else sent_state,
                                error=result.errors, sent_at=timezone.now(), extra_sender_data=extra_sender_data)
        except (JSONDecodeError, requests.exceptions.RequestException) as ex:
            self.update_message(message, state=PushNotificationMessage.STATE.ERROR, error=force_text(ex))
            # Do not re-raise caught exception. Re-raise exception causes transaction rollback (loss of information
            # about exception).



from chamber.models import SmartModel
from chamber.utils.datastructures import ChoicesNumEnum
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import Manager
from django.template import Template, Context
from django.template.exceptions import TemplateSyntaxError, TemplateDoesNotExist
from django.utils.translation import ugettext_lazy as _

from jsonfield.fields import JSONField

from pymess.utils import has_int_pk


class RelatedObjectManager(Manager):

    def create_from_related_object(self, related_object):
        return self.create(
            object_id_int=related_object.pk if has_int_pk(related_object) else None,
            object_id=related_object.pk,
            content_type=ContentType.objects.get_for_model(related_object)
        )

    def create_from_related_objects(self, *related_objects):
        return [
            self.create_from_related_object(related_object)
            for related_object in related_objects
        ]


class BaseMessage(SmartModel):

    STATE = ChoicesNumEnum()

    sent_at = models.DateTimeField(verbose_name=_('sent at'), null=True, blank=True, editable=False)
    recipient = models.CharField(verbose_name=_('recipient'), null=False, blank=False, max_length=20)
    content = models.TextField(verbose_name=_('content'), null=False, blank=False)
    template_slug = models.SlugField(verbose_name=_('slug'), max_length=100, null=True, blank=True, editable=False)
    state = models.IntegerField(verbose_name=_('state'), null=False, blank=False, choices=STATE.choices, editable=False)
    backend = models.CharField(verbose_name=_('backend'), null=True, blank=True, editable=False, max_length=250)
    error = models.TextField(verbose_name=_('error'), null=True, blank=True, editable=False)
    extra_data = JSONField(verbose_name=_('extra data'), null=True, blank=True, editable=False)
    extra_sender_data = JSONField(verbose_name=_('extra sender data'), null=True, blank=True, editable=False)
    tag = models.SlugField(verbose_name=_('tag'), null=True, blank=True, editable=False)

    def __str__(self):
        return self.recipient

    class Meta:
        abstract = True
        ordering = ('-created_at',)


class BaseRelatedObject(SmartModel):

    content_type = models.ForeignKey(ContentType, verbose_name=_('content type of the related object'),
                                     null=False, blank=False, on_delete=models.CASCADE)
    object_id = models.TextField(verbose_name=_('ID of the related object'), null=False, blank=False)
    object_id_int = models.PositiveIntegerField(verbose_name=_('ID of the related object in int format'), null=True,
                                                blank=True, db_index=True)
    content_object = GenericForeignKey('content_type', 'object_id')

    objects = RelatedObjectManager()

    class Meta:
        abstract = True
        ordering = ('-created_at',)


class BaseAbstractTemplate(SmartModel):

    slug = models.SlugField(verbose_name=_('slug'), max_length=100, null=False, blank=False, editable=False,
                            primary_key=True)
    body = models.TextField(verbose_name=_('message body'), null=True, blank=False)

    def _update_context_data(self, context_data):
        return context_data

    def render_text_template(self, text, context_data):
        context_data = self._update_context_data(context_data)
        return Template(text).render(Context(context_data))

    def render_body(self, context_data):
        return self.render_text_template(self.get_body(), context_data)

    def clean_body(self, context_data=None):
        try:
            self.render_body(context_data or {})
        except (TemplateSyntaxError, TemplateDoesNotExist) as ex:
            raise ValidationError(_('Error during template body rendering: "{}"').format(ex))

    def get_body(self):
        return self.body

    def can_send(self, recipient, context_data):
        return True

    def get_backend_sender(self):
        raise NotImplementedError

    def send(self, recipient, context_data, related_objects=None, tag=None, **kwargs):
        if self.can_send(recipient, context_data):
            return self.get_backend_sender().send(
                recipient=recipient,
                content=self.render_body(context_data),
                related_objects=related_objects,
                tag=tag,
                template=self,
                **kwargs,
            )
        else:
            return None

    def __str__(self):
        return self.slug

    class Meta:
        abstract = True
        ordering = ('-created_at',)



from django.db import models

from pymess.config import settings


def normalize_phone_number(number):
    """
    Function that normalize input phone number to the valid phone number format.
    """
    if number:
        number = number.replace(' ', '').replace('-', '')
        if len(number) == 9 and settings.SMS_DEFAULT_PHONE_CODE:
            number = ''.join((settings.SMS_DEFAULT_PHONE_CODE, number))
        elif len(number) == 14 and number.startswith('00'):
            number = '+' + number[2:]
    return number


def has_int_pk(model):
    """
    Tests whether the given model has an integer primary key.
    """
    pk = model._meta.pk
    return (
        (
            isinstance(pk, (models.IntegerField, models.AutoField)) and
            not isinstance(pk, models.BigIntegerField)
        ) or (
            isinstance(pk, models.ForeignKey) and has_int_pk(pk.rel.to)
        )
    )


def fullname(o):
    """
    Helper that returns name of the input object with its path.
    """
    return o.__module__ + "." + o.__class__.__name__



import logging
from datetime import timedelta

from django.utils.encoding import force_text
from django.utils.timezone import now

from chamber.exceptions import PersistenceException

from pymess.backend import BaseBackend, send_template as _send_template, send as _send
from pymess.config import settings, get_dialer_template_model, get_dialer_sender
from pymess.models import DialerMessage
from pymess.utils import fullname

LOGGER = logging.getLogger(__name__)


class DialerSendingError(Exception):
    pass


class DialerBackend(BaseBackend):
    """
    Base class for dialer backend containing implementation of concrete dialer service that
    is used for automatic call to selected phone number.
    """

    model = DialerMessage

    def create_message(self, recipient, content, related_objects, tag, template, **kwargs):
        """
        Create dialer message which will be logged in the database.
        :param recipient: phone number of the recipient
        :param content: content of the dialer message
        :param related_objects: list of related objects that will be linked with the dialer message using generic
        relation
        :param tag: string mark that will be saved with the message
        :param template: template object from which content of the message was created
        :param kwargs: extra attributes that will be saved with the message
        """
        try:
            return super().create_message(
                recipient,
                content,
                related_objects,
                tag,
                template,
                state=self.get_initial_dialer_state(recipient),
                extra_data=kwargs,
                **self._get_extra_message_kwargs()
            )
        except PersistenceException as ex:
            raise DialerSendingError(force_text(ex))

    def get_initial_dialer_state(self, recipient):
        """
        returns initial state for logged dialer instance.
        :param recipient: phone number of the recipient
        """
        return self.model.STATE.READY

    def _update_dialer_states(self, messages):
        """
        If dialer sender provides check dialer delivery this method can be overridden.
        :param messages: messages which state will be updated
        """
        raise NotImplementedError('Check dialer state is not supported with the backend')

    def bulk_check_dialer_status(self):
        """
        Method that finds messages that are not in the final state and updates their states.
        """
        messages_to_check = self.model.objects.filter(
            is_final_state=False,
            backend=fullname(self),
            created_at__gte=now() - timedelta(minutes=settings.DIALER_IDLE_MESSAGES_TIMEOUT_MINUTES),
        )
        if messages_to_check.exists():
            self._update_dialer_states(messages_to_check)


def send_template(recipient, slug, context_data, related_objects=None, tag=None):
    """
    Helper for building and sending dialer message from a template.
    :param recipient: phone number of the recipient
    :param slug: slug of a dialer template
    :param context_data: dict of data that will be sent to the template renderer
    :param related_objects: list of related objects that will be linked with the dialer message using generic
        relation
    :param tag: string mark that will be saved with the message
    :return: dialer message object or None if template cannot be sent
    """
    return _send_template(
        recipient,
        slug,
        context_data,
        related_objects,
        tag,
        template_model=get_dialer_template_model(),
    )


def send(recipient, content, related_objects=None, tag=None, **kwargs):
    """
    Helper for sending dialer message.
    :param recipient: phone number of the recipient
    :param content: text content of the messages
    :param related_objects:
    :param tag: string mark that will be saved with the message
    :param kwargs: extra attributes that will be stored with messages
    :return: True if dialer was successfully sent or False if message is in error state
    """
    return _send(
        recipient,
        content,
        related_objects,
        tag,
        message_sender=get_dialer_sender(),
        **kwargs
    )



def generate_session(slug=None, related_objects=None):
    try:
        from security.transport.security_requests import SecuritySession

        return SecuritySession(slug=slug, related_objects=related_objects)
    except ImportError:
        from requests import Session

        return Session()



from django.core.management.base import BaseCommand

from pymess.config import get_dialer_sender


class Command(BaseCommand):
    """
    Command for checking dialer call status. Dialer backend must support it.
    """

    def handle(self, *args, **kwargs):
        get_dialer_sender().bulk_check_dialer_status()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.test import TestCase

from pymess.backend.sms.sms_operator import SMSOperatorBackend


class OutputSMSTestCase(TestCase):

   def test_send(self):
       SMSOperatorBackend().send('+420604972057', 'test')

   def test_bulk_send(self):
       SMSOperatorBackend().bulk_send(['+420604972057', '+420604972058'], 'test')


from __future__ import unicode_literals

import string
from datetime import datetime, timedelta

import factory
from factory import django as factory_django
from factory import fuzzy

from sender.models import InputSMS, OutputSMS, SMSTemplate


class InputSMSFactory(factory_django.DjangoModelFactory):

    received_at = fuzzy.FuzzyNaiveDateTime(start_dt=datetime.now() - timedelta(days=150))
    uniq = factory.Sequence(lambda n: n)
    sender = '777123456'
    recipient = '777444888'
    okey = fuzzy.FuzzyText(length=50)
    opid = fuzzy.FuzzyText(length=50)
    opmid = fuzzy.FuzzyText(length=50)
    content = fuzzy.FuzzyText(length=50)
    created_at = fuzzy.FuzzyNaiveDateTime(start_dt=datetime.now() - timedelta(days=150))
    changed_at = fuzzy.FuzzyNaiveDateTime(start_dt=datetime.now() - timedelta(days=150))

    class Meta:
        model = InputSMS


class OutputSMSFactory(factory_django.DjangoModelFactory):
    # Do NOT add sent_at to the factory, must be set by hand to be able to test functionality
    sender = '+420777123456'
    recipient = fuzzy.FuzzyText(length=9, prefix='+420', chars=string.digits)
    opmid = ''
    dlr = '1'
    kw = fuzzy.FuzzyText(length=30)
    content = fuzzy.FuzzyText(length=160)
    created_at = fuzzy.FuzzyNaiveDateTime(start_dt=datetime.now() - timedelta(days=150))
    changed_at = fuzzy.FuzzyNaiveDateTime(start_dt=datetime.now() - timedelta(days=150))

    class Meta:
        model = OutputSMS


class SMSTemplateFactory(factory_django.DjangoModelFactory):

    slug = 'test'
    body = 'Does rendering context variables work? {{ variable }}'
    created_at = fuzzy.FuzzyNaiveDateTime(start_dt=datetime.now() - timedelta(days=150))
    changed_at = fuzzy.FuzzyNaiveDateTime(start_dt=datetime.now() - timedelta(days=150))

    class Meta:
        model = SMSTemplate



from chamber.shortcuts import change_and_save

from pymess.backend.sms import SMSBackend
from pymess.models import OutputSMSMessage


class DummySMSBackend(SMSBackend):
    """
    Dummy SMS backend used for testing environments. Backend only logs messages to the database.
    """

    def publish_message(self, message):
        self.update_message(message, state=OutputSMSMessage.STATE.DEBUG)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.16 on 2019-02-26 11:39
from __future__ import unicode_literals

import jsonfield.fields

import django.db.models.deletion
from django.db import migrations, models
from pymess.config import settings


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('pymess', '0004_dialermessage_is_final_state'),
    ]

    operations = [
        migrations.CreateModel(
            name='PushNotificationMessage',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at')),
                ('sent_at', models.DateTimeField(blank=True, editable=False, null=True, verbose_name='sent at')),
                ('recipient', models.CharField(max_length=20, verbose_name='recipient')),
                ('content', models.TextField(verbose_name='content')),
                ('template_slug', models.SlugField(blank=True, editable=False, max_length=100, null=True,
                                                   verbose_name='slug')),
                ('backend', models.CharField(blank=True, editable=False, max_length=250, null=True,
                                             verbose_name='backend')),
                ('error', models.TextField(blank=True, editable=False, null=True, verbose_name='error')),
                ('extra_data', jsonfield.fields.JSONField(blank=True, editable=False, null=True,
                                                          verbose_name='extra data')),
                ('extra_sender_data', jsonfield.fields.JSONField(blank=True, editable=False, null=True,
                                                                 verbose_name='extra sender data')),
                ('tag', models.SlugField(blank=True, editable=False, null=True, verbose_name='tag')),
                ('state', models.PositiveIntegerField(choices=[(1, 'waiting'), (2, 'sent'), (3, 'error'), (4, 'debug')],
                                                      editable=False, verbose_name='state')),
                ('heading', models.TextField(verbose_name='heading')),
                ('url', models.URLField(null=True, blank=True, verbose_name='URL')),
            ],
            options={
                'verbose_name': 'push notification',
                'verbose_name_plural': 'push notifications',
                'ordering': ('-created_at',),
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='PushNotificationMessageRelatedObject',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at')),
                ('object_id', models.TextField(verbose_name='ID of the related object')),
                ('object_id_int', models.PositiveIntegerField(blank=True, db_index=True, null=True,
                                                              verbose_name='ID of the related object in int format')),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE,
                                                   to='contenttypes.ContentType',
                                                   verbose_name='content type of the related object')),
                ('push_notification_message', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE,
                                                                related_name='related_objects',
                                                                to='pymess.PushNotificationMessage',
                                                                verbose_name='push notification message')),
            ],
            options={
                'verbose_name': 'related object of a push notification message',
                'verbose_name_plural': 'related objects of a push notification message',
                'ordering': ('-created_at',),
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='PushNotificationTemplate',
            fields=[
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at')),
                ('slug', models.SlugField(editable=False, max_length=100, primary_key=True, serialize=False,
                                          verbose_name='slug')),
                ('body', models.TextField(null=True, verbose_name='message body')),
                ('heading', models.TextField(verbose_name='heading')),
                ('is_active', models.BooleanField(default=True, verbose_name='is_active')),
            ],
            options={
                'verbose_name': 'push notification template',
                'verbose_name_plural': 'push notification templates',
                'ordering': ('-created_at',),
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='pushnotificationmessage',
            name='template',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL,
                                    related_name='push_notifications', to=settings.PUSH_NOTIFICATION_TEMPLATE_MODEL,
                                    verbose_name='template'),
        ),
    ]



from django.utils.encoding import force_text

from chamber.exceptions import PersistenceException

from pymess.backend import BaseBackend, send_template as _send_template, send as _send
from pymess.config import get_email_template_model, get_email_sender, settings
from pymess.models import EmailMessage


class EmailBackend(BaseBackend):
    """
    Base class for E-mail backend containing implementation of e-mail service that
    is used for sending messages.
    """

    model = EmailMessage

    class EmailSendingError(Exception):
        pass

    def update_message(self, message, extra_sender_data=None, **kwargs):
        """
        Method for updating state of the message after sending
        :param message: e-mail message object
        :param extra_sender_data: extra data that will be saved to the extra_sender_data field
        :param kwargs: changed object kwargs
        :return:
        """
        super().update_message(
            message,
            extra_sender_data=extra_sender_data,
            number_of_send_attempts=message.number_of_send_attempts + 1,
            **kwargs
        )

    def create_message(self, sender, sender_name, recipient, subject, content, related_objects, tag, template,
                       attachments, **kwargs):
        """
        Create e-mail which will be logged in the database.
        :param sender: e-mail address of the sender
        :param sender_name: friendly name of the sender
        :param recipient: e-mail address of the receiver
        :param subject: subject of the e-mail message
        :param content: content of the e-mail message
        :param related_objects: list of related objects that will be linked with the e-mail message using generic
        relation
        :param tag: string mark that will be saved with the message
        :param template: template object from which content, subject and sender of the message was created
        :param attachments: list of files that will be sent with the message as attachments
        :param kwargs: extra data that will be saved in JSON format in the extra_data model field
        """
        try:
            message = super().create_message(
                recipient=recipient,
                content=content,
                related_objects=related_objects,
                tag=tag,
                template=template,
                sender=sender,
                sender_name=sender_name,
                subject=subject,
                state=self.get_initial_email_state(recipient),
                extra_data=kwargs,
                **self._get_extra_message_kwargs()
            )
            if attachments:
                message.attachments.create_from_tripples(*attachments)
            return message
        except PersistenceException as ex:
            raise self.EmailSendingError(force_text(ex))

    def send(self, sender, recipient, subject, content, sender_name=None, related_objects=None, tag=None,
             template=None, attachments=None, **kwargs):
        """
        Send e-mail with defined values
        :param sender: e-mail address of the sender
        :param recipient: e-mail address of the receiver
        :param subject: subject of the e-mail message
        :param content: content of the e-mail message
        :param sender_name: friendly name of the sender
        :param related_objects: list of related objects that will be linked with the e-mail message with generic
            relation
        :param tag: string mark that will be saved with the message
        :param template: template object from which content, subject and sender was created
        :param attachments: list of files that will be sent with the message as attachments
        :param kwargs: extra data that will be saved in JSON format in the extra_data model field
        """
        message = self.create_message(
            sender, sender_name, recipient, subject, content, related_objects, tag, template, attachments, **kwargs
        )
        if not settings.EMAIL_BATCH_SENDING:
            self.publish_message(message)
        return message

    def get_initial_email_state(self, recipient):
        """
        returns initial state for logged e-mail message.
        :param recipient: e-mail address of the recipient
        """
        return self.model.STATE.WAITING


def send_template(recipient, slug, context_data, related_objects=None, attachments=None, tag=None):
    """
    Helper for building and sending e-mail message from a template.
    :param recipient: e-mail address of the receiver
    :param slug: slug of the e-mail template
    :param context_data: dict of data that will be sent to the template renderer
    :param related_objects: list of related objects that will be linked with the e-mail message with generic
        relation
    :param attachments: list of files that will be sent with the message as attachments
    :param tag: string mark that will be saved with the message
    :return: e-mail message object or None if template cannot be sent
    """
    return _send_template(
        recipient=recipient,
        slug=slug,
        context_data=context_data,
        related_objects=related_objects,
        tag=tag,
        template_model=get_email_template_model(),
        attachments=attachments,
    )


def send(sender, recipient, subject, content, sender_name=None, related_objects=None, attachments=None, tag=None,
         **email_kwargs):
    """
    Helper for sending e-mail message.
    :param sender: e-mail address of the sender
    :param recipient: e-mail address of the receiver
    :param subject: subject of the e-mail message
    :param content: content of the e-mail message
    :param sender_name: friendly name of the sender
    :param related_objects: list of related objects that will be linked with the e-mail message with generic
        relation
    :param tag: string mark that will be saved with the message
    :param attachments: list of files that will be sent with the message as attachments
    :param email_kwargs: extra data that will be saved in JSON format in the extra_data model field
    :return: True if e-mail was successfully sent or False if e-mail is in error state
    """
    return get_email_sender().send(
        sender=sender,
        recipient=recipient,
        subject=subject,
        content=content,
        sender_name=sender_name,
        related_objects=related_objects,
        tag=tag,
        attachments=attachments,
        **email_kwargs
    ).failed



import os
import base64

import requests

from json.decoder import JSONDecodeError

from django.utils import timezone
from django.utils.encoding import force_text
from django.utils.translation import ugettext_lazy as _

from chamber.utils.datastructures import ChoicesEnum

import mandrill

from pymess.backend.emails import EmailBackend
from pymess.models import EmailMessage
from pymess.config import settings
from pymess.utils.logged_requests import generate_session


class MandrillEmailBackend(EmailBackend):
    """
    E-mail backend implementing Mandrill service (https://mandrillapp.com/api/docs/index.python.html).
    """

    MANDRILL_STATES = ChoicesEnum(
        ('SENT', _('sent')),
        ('QUEUED', _('queued')),
        ('SCHEDULED', _('scheduled')),
        ('REJECTED', _('rejected')),
        ('INVALID', _('invalid')),
    )

    MANDRILL_STATES_MAPPING = {
        MANDRILL_STATES.SENT: EmailMessage.STATE.SENT,
        MANDRILL_STATES.QUEUED: EmailMessage.STATE.SENT,
        MANDRILL_STATES.SCHEDULED: EmailMessage.STATE.SENT,
        MANDRILL_STATES.REJECTED: EmailMessage.STATE.ERROR,
        MANDRILL_STATES.INVALID: EmailMessage.STATE.ERROR,
    }

    def _serialize_attachments(self, message):
        return [
            {
                'type': attachment.content_type,
                'name': os.path.basename(attachment.file.name),
                'content': base64.b64encode(attachment.file.read()).decode('utf-8')
            } for attachment in message.attachments.all()
        ]

    def publish_message(self, message):
        mandrill_client = mandrill.Mandrill(settings.EMAIL_MANDRILL.KEY)
        mandrill_client.session = generate_session(slug='pymess - Mandrill', related_objects=(message,))
        try:
            result = mandrill_client.messages.send(
                message={
                    'to': [{'email': message.recipient}],
                    'from_email': message.sender,
                    'from_name': message.sender_name,
                    'html': message.content,
                    'subject': message.subject,
                    'headers': settings.EMAIL_MANDRILL.HEADERS,
                    'track_opens': settings.EMAIL_MANDRILL.TRACK_OPENS,
                    'auto_text': settings.EMAIL_MANDRILL.AUTO_TEXT,
                    'inline_css': settings.EMAIL_MANDRILL.INLINE_CSS,
                    'url_strip_qs': settings.EMAIL_MANDRILL.URL_STRIP_QS,
                    'preserve_recipients': settings.EMAIL_MANDRILL.PRESERVE_RECIPIENTS,
                    'view_content_link': settings.EMAIL_MANDRILL.VIEW_CONTENT_LINK,
                    'async': settings.EMAIL_MANDRILL.ASYNC,
                    'attachments': self._serialize_attachments(message)
                },
            )[0]
            mandrill_state = result['status'].upper()
            state = self.MANDRILL_STATES_MAPPING.get(mandrill_state)
            error = self.MANDRILL_STATES.get_label(mandrill_state) if state == EmailMessage.STATE.ERROR else None
            if mandrill_state == self.MANDRILL_STATES.REJECTED:
                error += ', mandrill message: "{}"'.format(result['reject_reason'])

            extra_sender_data = message.extra_sender_data or {}
            extra_sender_data['result'] = result
            self.update_message(message, state=state, sent_at=timezone.now(),
                                extra_sender_data=extra_sender_data, error=error)
        except (mandrill.Error, JSONDecodeError, requests.exceptions.RequestException) as ex:
            self.update_message(message, state=EmailMessage.STATE.ERROR, error=force_text(ex))
            # Do not re-raise caught exception. Re-raise exception causes transaction rollback (lost of information
            # about exception).



from django.core.management.base import BaseCommand

from pymess.config import get_sms_sender


class Command(BaseCommand):
    """
    Command for checking SMS delivery. SMS backend must support it.
    """

    def handle(self, *args, **kwargs):
        get_sms_sender().bulk_check_sms_states()



from bs4 import BeautifulSoup

import requests

from django.utils import timezone
from django.utils.encoding import force_text
from django.utils.translation import ugettext_lazy as _
from django.utils.safestring import mark_safe
from django.template.loader import render_to_string

from chamber.utils.datastructures import ChoicesNumEnum, Enum

from pymess.backend.sms import SMSBackend
from pymess.models import OutputSMSMessage
from pymess.utils.logged_requests import generate_session
from pymess.config import settings


class ATSSMSBackend(SMSBackend):
    """
    SMS backend that implements ATS operator service https://www.atspraha.cz/
    Backend supports check SMS delivery
    """

    REQUEST_TYPES = Enum(
        'SMS',
        'DELIVERY_REQUEST',
    )

    TEMPLATES = {
        'base': 'pymess/sms/ats/base.xml',
        REQUEST_TYPES.SMS: 'pymess/sms/ats/sms.xml',
        REQUEST_TYPES.DELIVERY_REQUEST: 'pymess/sms/ats/delivery_request.xml',
    }

    class ATSSendingError(Exception):
        pass

    ATS_STATES = ChoicesNumEnum(
        # SMS delivery receipts
        ('NOT_FOUND', _('not found'), 20),
        ('NOT_SENT', _('not sent yet'), 21),
        ('SENT', _('sent'), 22),
        ('DELIVERED', _('delivered'), 23),
        ('NOT_DELIVERED', _('not delivered'), 24),
        ('UNKNOWN', _('not able to determine the state'), 25),
        # Authentication
        ('AUTHENTICATION_FAILED', _('authentication failed'), 100),
        # Internal errors
        ('DB_ERROR', _('DB error'), 200),
        # Request states
        ('OK', _('SMS is OK and ready to be sent'), 0),
        ('UNSPECIFIED_ERROR', _('unspecified error'), 1),
        ('BATCH_WITH_NOT_UNIQUE_UNIQ', _('one of the requests has not unique "uniq"'), 300),
        ('SMS_NOT_UNIQUE_UNIQ', _('SMS has not unique "uniq"'), 310),
        ('SMS_NO_KW', _('SMS lacks keyword'), 320),
        ('KW_INVALID', _('keyword not valid'), 321),
        ('NO_SENDER', _('no sender specified'), 330),
        ('SENDER_INVALID', _('sender not valid'), 331),
        ('MO_PR_NOT_ALLOWED', _('MO PR SMS not allowed'), 332),
        ('MT_PR_NOT_ALLOWED', _('MT PR SMS not allowed'), 333),
        ('MT_PR_DAILY_LIMIT', _('MT PR SMS daily limit exceeded'), 334),
        ('MT_PR_TOTAL_LIMIT', _('MT PR SMS total limit exceeded'), 335),
        ('GEOGRAPHIC_NOT_ALLOWED', _('geographic number is not allowed'), 336),
        ('MT_SK_NOT_ALLOWED', _('MT SMS to Slovakia not allowed'), 337),
        ('SHORTCODES_NOT_ALLOWED', _('shortcodes not allowed'), 338),
        ('UNKNOWN_SENDER', _('sender is unknown'), 339),
        ('UNSPECIFIED_SMS_TYPE', _('type of SMS not specified'), 340),
        ('TOO_LONG', _('SMS too long'), 341),
        ('TOO_MANY_PARTS', _('too many SMS parts (max. is 10)'), 342),
        ('WRONG_SENDER_OR_RECEIVER', _('wrong number of sender/receiver'), 343),
        ('NO_RECIPIENT_OR_WRONG_FORMAT', _('recipient is missing or in wrong format'), 350),
        ('TEXTID_NOT_ALLOWED', _('using "textid" is not allowed'), 360),
        ('WRONG_TEXTID', _('"textid" is in wrong format'), 361),
        ('LONG_SMS_TEXTID_NOT_ALLOWED', _('long SMS with "textid" not allowed'), 362),
        # XML errors
        ('XML_MISSING', _('XML body missing'), 701),
        ('XML_UNREADABLE', _('XML is not readable'), 702),
        ('WRONG_HTTP_METHOD', _('unknown HTTP method or not HTTP POST'), 703),
        ('XML_INVALID', _('XML invalid'), 705),
    )

    ATS_STATES_MAPPING = {
        ATS_STATES.NOT_FOUND: OutputSMSMessage.STATE.ERROR,
        ATS_STATES.NOT_SENT: OutputSMSMessage.STATE.SENDING,
        ATS_STATES.SENT: OutputSMSMessage.STATE.SENT,
        ATS_STATES.DELIVERED: OutputSMSMessage.STATE.DELIVERED,
        ATS_STATES.NOT_DELIVERED: OutputSMSMessage.STATE.ERROR,
        ATS_STATES.OK: OutputSMSMessage.STATE.SENDING,
        ATS_STATES.UNSPECIFIED_ERROR: OutputSMSMessage.STATE.ERROR,
        ATS_STATES.BATCH_WITH_NOT_UNIQUE_UNIQ: OutputSMSMessage.STATE.ERROR,
        ATS_STATES.SMS_NOT_UNIQUE_UNIQ: OutputSMSMessage.STATE.ERROR,
        ATS_STATES.SMS_NO_KW: OutputSMSMessage.STATE.ERROR,
        ATS_STATES.KW_INVALID: OutputSMSMessage.STATE.ERROR,
        ATS_STATES.NO_SENDER: OutputSMSMessage.STATE.ERROR,
        ATS_STATES.SENDER_INVALID: OutputSMSMessage.STATE.ERROR,
        ATS_STATES.MO_PR_NOT_ALLOWED: OutputSMSMessage.STATE.ERROR,
        ATS_STATES.MT_SK_NOT_ALLOWED: OutputSMSMessage.STATE.ERROR,
        ATS_STATES.SHORTCODES_NOT_ALLOWED: OutputSMSMessage.STATE.ERROR,
        ATS_STATES.UNKNOWN_SENDER: OutputSMSMessage.STATE.ERROR,
        ATS_STATES.UNSPECIFIED_SMS_TYPE: OutputSMSMessage.STATE.ERROR,
        ATS_STATES.TOO_LONG: OutputSMSMessage.STATE.ERROR,
        ATS_STATES.TOO_MANY_PARTS: OutputSMSMessage.STATE.ERROR,
        ATS_STATES.WRONG_SENDER_OR_RECEIVER: OutputSMSMessage.STATE.ERROR,
        ATS_STATES.NO_RECIPIENT_OR_WRONG_FORMAT: OutputSMSMessage.STATE.ERROR,
        ATS_STATES.TEXTID_NOT_ALLOWED: OutputSMSMessage.STATE.ERROR,
        ATS_STATES.WRONG_TEXTID: OutputSMSMessage.STATE.ERROR,
        ATS_STATES.LONG_SMS_TEXTID_NOT_ALLOWED: OutputSMSMessage.STATE.ERROR,
    }

    def __init__(self):
        self.config = settings.ATS_SMS_CONFIG

    def _get_extra_sender_data(self):
        return {
            'prefix': self.config.UNIQ_PREFIX,
            'validity': self.config.VALIDITY,
            'kw': self.config.PROJECT_KEYWORD,
            'textid': self.config.TEXTID,
        }

    def _get_extra_message_kwargs(self):
        return {
            'sender': self.config.OUTPUT_SENDER_NUMBER,
        }

    def _serialize_messages(self, messages, request_type):
        """
        Serialize SMS messages to the XML
        :param messages: list of SMS messages
        :param request_type: type of the request to the ATS operator
        :return: serialized XML message that will be sent to the ATS service
        """
        return render_to_string(
            self.TEMPLATES['base'], {
                'username': self.config.USERNAME,
                'password': self.config.PASSWORD,
                'template_type': self.TEMPLATES[request_type],
                'messages': messages,
                'prefix': str(self.config.UNIQ_PREFIX) + '-',
                'sender': self.config.OUTPUT_SENDER_NUMBER,
                'dlr': 1,
                'validity': self.config.VALIDITY,
                'kw': self.config.PROJECT_KEYWORD,
                'billing': 0,
                'extra': mark_safe(' textid="{textid}"'.format(
                    textid=self.config.TEXTID
                )) if self.config.TEXTID else '',
            }
        )

    def _send_requests(self, messages, request_type, **change_sms_kwargs):
        """
        Performs the actual POST request for input messages and request type.
        :param messages: list of SMS messages
        :param request_type: type of the request
        :param change_sms_kwargs: extra kwargs that will be stored to the message object
        """
        requests_xml = self._serialize_messages(messages, request_type)
        try:
            resp = generate_session(slug='pymess - ATS SMS', related_objects=list(messages)).post(
                self.config.URL,
                data=requests_xml,
                headers={'Content-Type': 'text/xml'}
            )
            if resp.status_code != 200:
                raise self.ATSSendingError(
                    'ATS operator returned invalid response status code: {}'.format(resp.status_code)
                )
            self._update_sms_states_from_response(messages, self._parse_response_codes(resp.text), **change_sms_kwargs)
        except requests.exceptions.RequestException as ex:
            raise self.ATSSendingError(
                'ATS operator returned returned exception: {}'.format(force_text(ex))
            )

    def _update_sms_states_from_response(self, messages, parsed_response, **change_sms_kwargs):
        """
        Higher-level function performing serialization of ATS requests, parsing ATS server response and updating
        SMS messages state according the received response.
        :param messages: list of SMS messages
        :param parsed_response: parsed HTTP response from the ATS service
        :param change_sms_kwargs: extra kwargs that will be stored to the message object
        """

        messages_dict = {message.pk: message for message in messages}

        missing_uniq = set(messages_dict.keys()) - set(parsed_response.keys())
        if missing_uniq:
            raise self.ATSSendingError(
                'ATS operator not returned SMS info with uniq: {}'.format(', '.join(map(str, missing_uniq)))
            )

        extra_uniq = set(parsed_response.keys()) - set(messages_dict.keys())
        if extra_uniq:
            raise self.ATSSendingError(
                'ATS operator returned SMS info about unknown uniq: {}'.format(', '.join(map(str, extra_uniq)))
            )

        for uniq, ats_state in parsed_response.items():
            sms = messages_dict[uniq]
            state = self.ATS_STATES_MAPPING.get(ats_state)
            error = self.ATS_STATES.get_label(ats_state) if state == OutputSMSMessage.STATE.ERROR else None
            self.update_message(
                sms,
                state=state,
                error=error,
                extra_sender_data={'sender_state': ats_state}
                **change_sms_kwargs
            )

    def publish_messages(self, messages):
        self._send_requests(messages, request_type=self.REQUEST_TYPES.SMS, sent_at=timezone.now())

    def publish_message(self, message):
        self._send_requests([message], request_type=self.REQUEST_TYPES.SMS, sent_at=timezone.now())

    def _parse_response_codes(self, xml):
        """
        Finds all <code> tags in the given XML and returns a mapping "uniq" -> "response code" for all SMS.
        In case of an error, the error is logged.
        :param xml: XML from the ATL response
        :return: dictionary with pair {SMS uniq: response status code}
        """

        soup = BeautifulSoup(xml, 'html.parser')
        code_tags = soup.find_all('code')

        error_message = ', '.join(
            [(force_text(self.ATS_STATES.get_label(c))
              if c in self.ATS_STATES.all
              else 'ATS returned an unknown state {}.'.format(c))
             for c in [int(error_code.string) for error_code in code_tags if not error_code.attrs.get('uniq')]],
        )

        if error_message:
            raise self.ATSSendingError('Error returned from ATS operator: {}'.format(error_message))

        return {
            int(code.attrs['uniq'].lstrip(str(self.config.UNIQ_PREFIX) + '-')): int(code.string)
            for code in code_tags if code.attrs.get('uniq')
        }

    def _update_sms_states(self, messages):
        self._send_requests(messages, request_type=self.REQUEST_TYPES.DELIVERY_REQUEST)



from django.apps import apps
from django.conf import settings as django_settings
from django.utils.module_loading import import_string

from attrdict import AttrDict


DEFAULTS = {
    # SMS configuration
    'SMS_TEMPLATE_MODEL': 'pymess.SMSTemplate',
    'SMS_ATS_CONFIG': {
        'UNIQ_PREFIX': '',
        'VALIDITY': 60,
        'TEXTID': None,
        'URL': 'http://fik.atspraha.cz/gwfcgi/XMLServerWrapper.fcgi',
        'OPTID': '',
    },
    'SMS_OPERATOR_CONFIG': {
        'URL': 'https://www.sms-operator.cz/webservices/webservice.aspx',
        'UNIQ_PREFIX': '',
    },
    'SMS_SNS_CONFIG': {
    },
    'SMS_USE_ACCENT': False,
    'SMS_DEFAULT_PHONE_CODE': None,
    'SMS_SENDER_BACKEND': 'pymess.backend.sms.dummy.DummySMSBackend',
    'SMS_LOG_IDLE_MESSAGES': True,
    'SMS_SET_ERROR_TO_IDLE_MESSAGES': True,
    'SMS_IDLE_MESSAGES_TIMEOUT_MINUTES': 10,

    # E-mail configuration
    'EMAIL_TEMPLATE_MODEL': 'pymess.EmailTemplate',
    'EMAIL_SENDER_BACKEND': 'pymess.backend.emails.dummy.DummyEmailBackend',
    'EMAIL_MANDRILL': {
        'HEADERS': None,
        'TRACK_OPENS': False,
        'TRACK_CLICKS': False,
        'AUTO_TEXT': False,
        'INLINE_CSS': False,
        'URL_STRIP_QS': False,
        'PRESERVE_RECIPIENTS': False,
        'VIEW_CONTENT_LINK': True,
        'ASYNC': False,
    },
    'EMAIL_BATCH_SENDING': False,
    'EMAIL_BATCH_LOCK_FILE': 'pymess_send_batch_emails',
    'EMAIL_BATCH_LOCK_WAIT_TIMEOUT': -1,
    'EMAIL_BATCH_SIZE': 20,
    'EMAIL_BATCH_MAX_NUMBER_OF_SEND_ATTEMPTS': 3,
    'EMAIL_BATCH_MAX_SECONDS_TO_SEND': 60 * 60,
    'EMAIL_SENDERS': (),
    'EMAIL_HTML_DATA_DIRECTORY': None,

    # Dialer configuration
    'DIALER_TEMPLATE_MODEL': 'pymess.DialerTemplate',
    'DIALER_SENDER_BACKEND': 'pymess.backend.dialer.dummy.DummyDialerBackend',
    'DIALER_DAKTELA': {
        'ACCESS_TOKEN':  None,
        'QUEUE': None,
        'URL': None,
        'STATES_MAPPING': {
            '0': 0,
            '1': 1,
            '2': 2,
            '3': 3,
            '4': 4,
            '5': 5,
            '6': 6,
        },
    },
    'DIALER_IDLE_MESSAGES_TIMEOUT_MINUTES': 60 * 24,

    # Push notification settings
    'PUSH_NOTIFICATION_TEMPLATE_MODEL': 'pymess.PushNotificationTemplate',
    'PUSH_NOTIFICATION_SENDER_BACKEND': 'pymess.backend.push.dummy.DummyPushNotificationBackend',
    'PUSH_NOTIFICATION_ONESIGNAL_APP_ID': None,
    'PUSH_NOTIFICATION_ONESIGNAL_API_KEY': None,
    'PUSH_NOTIFICATION_ONESIGNAL_LANGUAGE': None,
}


class Settings(object):
    """
    Pymess settings is loaded lazy like Django settings.
    The reason is usability of override_settings decorator with tests
    """

    def __getattr__(self, attr):
        if attr not in DEFAULTS:
            raise AttributeError('Invalid Pymess setting: "{}"'.format(attr))

        default_value = DEFAULTS[attr]
        value = getattr(django_settings, 'PYMESS_{}'.format(attr), DEFAULTS[attr])

        if isinstance(default_value, dict) and isinstance(value, dict):
            default_value = default_value.copy()
            default_value.update(value)
            value = AttrDict(default_value)

        return value


settings = Settings()


def get_model(model_name):
    """
    Helper that returns django model class defined by string {app_label}.{model_name}
    """
    return apps.get_model(*model_name.split('.'))


def get_sms_template_model():
    """
    Function returns SMS template model defined in Pymess settings
    """
    return get_model(settings.SMS_TEMPLATE_MODEL)


def get_email_template_model():
    """
    Function returns e-mail template model defined in Pymess settings
    """
    return get_model(settings.EMAIL_TEMPLATE_MODEL)


def get_sms_sender():
    """
    Function returns SMS sender backend from string defined in Pymess settings
    """
    return import_string(settings.SMS_SENDER_BACKEND)()


def get_email_sender():
    """
    Function returns e-mail sender backend from string defined in Pymess settings
    """
    return import_string(settings.EMAIL_SENDER_BACKEND)()


def get_dialer_template_model():
    """
    Function returns dialer template model defined in Pymess settings
    """
    return get_model(settings.DIALER_TEMPLATE_MODEL)


def get_dialer_sender():
    """
    Function returns dialer sender backend from string defined in Pymess settings
    """
    return import_string(settings.DIALER_SENDER_BACKEND)()


def get_push_notification_template_model():
    """
    Function returns push notification template model defined in Pymess settings
    """
    return get_model(settings.PUSH_NOTIFICATION_TEMPLATE_MODEL)()


def get_push_notification_sender():
    """
    Function returns push notification sender backend from string defined in Pymess settings
    """
    return import_string(settings.PUSH_NOTIFICATION_SENDER_BACKEND)()



import logging

from datetime import timedelta

from django.utils import timezone
from django.utils.encoding import force_text
from django.utils.translation import ugettext

from chamber.exceptions import PersistenceException

from pymess.backend import BaseBackend, send_template as _send_template, send as _send
from pymess.config import settings, get_sms_template_model, get_sms_sender
from pymess.models import OutputSMSMessage
from pymess.utils import fullname


LOGGER = logging.getLogger(__name__)


class SMSBackend(BaseBackend):
    """
    Base class for SMS backend containing implementation of concrete SMS service that
    is used for sending SMS messages.
    """

    model = OutputSMSMessage

    class SMSSendingError(Exception):
        pass

    def create_message(self, recipient, content, related_objects, tag, template, **kwargs):
        """
        Create SMS which will be logged in the database.
        :param recipient: phone number of the recipient
        :param content: content of the SMS message
        :param related_objects: list of related objects that will be linked with the SMS message using generic
        relation
        :param tag: string mark that will be saved with the message
        :param template: template object from which content of the message was created
        :param kwargs: extra attributes that will be saved with the message
        """
        try:
            return super().create_message(
                recipient=recipient,
                content=content,
                related_objects=related_objects,
                tag=tag,
                template=template,
                state=self.get_initial_sms_state(recipient),
                extra_data=kwargs,
                **self._get_extra_message_kwargs()
            )
        except PersistenceException as ex:
            raise self.SMSSendingError(force_text(ex))

    def get_initial_sms_state(self, recipient):
        """
        returns initial state for logged SMS instance.
        :param recipient: phone number of the recipient
        """
        return self.model.STATE.WAITING

    def _update_sms_states(self, messages):
        """
        If SMS sender provides check SMS delivery this method can be overridden.
        :param messages: messages which state will be updated
        """
        raise NotImplementedError('Check SMS state is not supported with the backend')

    def bulk_check_sms_states(self):
        """
        Method that find messages that is not in the final state and updates its states.
        """
        messages_to_check = self.model.objects.filter(state=self.model.STATE.SENDING)
        if messages_to_check.exists():
            self._update_sms_states(messages_to_check)

        idle_output_sms = self.model.objects.filter(
            state=self.model.STATE.SENDING,
            created_at__lt=timezone.now() - timedelta(minutes=settings.SMS_IDLE_MESSAGES_TIMEOUT_MINUTES),
            backend=fullname(self)
        )
        if settings.SMS_LOG_IDLE_MESSAGES and idle_output_sms.exists():
            LOGGER.warning('{count_sms} Output SMS is more than {timeout} minutes in state "SENDING"'.format(
                count_sms=idle_output_sms.count(), timeout=settings.SMS_IDLE_MESSAGES_TIMEOUT_MINUTES
            ))

        if settings.SMS_SET_ERROR_TO_IDLE_MESSAGES:
            idle_output_sms.update(
                state=self.model.STATE.ERROR, error=ugettext('timeouted')
            )


def send_template(recipient, slug, context_data, related_objects=None, tag=None):
    """
    Helper for building and sending SMS message from a template.
    :param recipient: phone number of the recipient
    :param slug: slug of a SMS template
    :param context_data: dict of data that will be sent to the template renderer
    :param related_objects: list of related objects that will be linked with the SMS message using generic
        relation
    :param tag: string mark that will be saved with the message
    :return: SMS message object or None if template cannot be sent
    """
    return _send_template(
        recipient=recipient,
        slug=slug,
        context_data=context_data,
        related_objects=related_objects,
        tag=tag,
        template_model=get_sms_template_model(),
    )


def send(recipient, content, related_objects=None, tag=None, **kwargs):
    """
    Helper for sending SMS message.
    :param recipient: phone number of the recipient
    :param content: text content of the messages
    :param related_objects:
    :param tag: string mark that will be saved with the message
    :param kwargs: extra attributes that will be stored with messages
    :return: True if SMS was successfully sent or False if message is in error state
    """
    return _send(
        recipient=recipient,
        content=content,
        related_objects=related_objects,
        tag=tag,
        message_sender=get_sms_sender(),
        **kwargs
    )
